/*
 * Copyright (c) 2021, 2023 Arm Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#if defined(__aarch64__)

#include "arm_gemm.hpp"
#include <cstddef>
#include <cstdint>

namespace arm_conv {
namespace depthwise {

void a64_u8q_packed_to_nhwc_5x5_s1_with_multiplier_output4x2_dot_depthfirst_impl(
  const uint8_t *const *const inptrs,
  uint8_t *const *const outptrs,
  const void *params,
  unsigned int n_output_channels,
  const arm_gemm::Requantize32& qp
)
{
  __asm__ __volatile__(
    "ldr q12, [%x[params], #0x0]\n"
    "ldr q8, [%x[params], #0x10]\n"
    "movi v28.16b, #0x1\n"
    "movi v18.4s, #0x0\n"
    "ldr q9, [%x[params], #0x20]\n"
    "ldr q10, [%x[params], #0x30]\n"
    "movi v31.4s, #0x0\n"
    "movi v24.4s, #0x0\n"
    "ldr q11, [%x[params], #0x40]\n"
    "ldr x20, [%x[inptrs], #0x18]\n"
    "movi v30.4s, #0x0\n"
    "movi v21.4s, #0x0\n"
    "ld1 { v3.16b }, [x20]\n"
    "ldr x20, [%x[inptrs], #0x20]\n"
    "mov v16.16b, v3.16b\n"
    "ext v16.16b, v16.16b, v16.16b, #0x1\n"
    "ld1 { v4.16b }, [x20]\n"
    "ldr x20, [%x[inptrs], #0x10]\n"
    "mov v15.16b, v4.16b\n"
    "ext v15.16b, v15.16b, v15.16b, #0x1\n"
    "ld1 { v2.16b }, [x20]\n"
    "ldr x20, [%x[inptrs], #0x8]\n"
    "mov v20.16b, v2.16b\n"
    "ext v20.16b, v20.16b, v20.16b, #0x1\n"
    "ld1 { v1.16b }, [x20]\n"
    "ldr x20, [%x[inptrs], #0x28]\n"
    "zip1 v3.2d, v3.2d, v16.2d\n"
    "zip1 v4.2d, v4.2d, v15.2d\n"
    "ld1 { v5.16b }, [x20]\n"
    "ldr x20, [%x[inptrs], #0x30]\n"
    "mov v26.16b, v1.16b\n"
    "mov v13.16b, v5.16b\n"
    "ld1 { v6.16b }, [x20]\n"
    "ldr x20, [%x[inptrs], #0x38]\n"
    "mov v19.16b, v6.16b\n"
    "ext v26.16b, v26.16b, v26.16b, #0x1\n"
    "ld1 { v7.16b }, [x20]\n"
    "ldr x20, [%x[inptrs], #0x0]\n"
    "mov v17.16b, v7.16b\n"
    "zip1 v2.2d, v2.2d, v20.2d\n"
    "ld1 { v0.16b }, [x20]\n"
    "ext v13.16b, v13.16b, v13.16b, #0x1\n"
    "ext v19.16b, v19.16b, v19.16b, #0x1\n"
    ".inst 0x6f83e392  // udot v18.4s, v28.16b, v3.4b[0]\n"
    "ext v17.16b, v17.16b, v17.16b, #0x1\n"
    ".inst 0x6f83eb9f  // udot v31.4s, v28.16b, v3.4b[2]\n"
    ".inst 0x6f84e398  // udot v24.4s, v28.16b, v4.4b[0]\n"
    "add x20, %x[qp], %[offsetof_Requantize32_b_offset]\n"
    "ld1r { v23.4s }, [x20]\n"
    ".inst 0x6f84eb9e  // udot v30.4s, v28.16b, v4.4b[2]\n"
    "mov v16.16b, v0.16b\n"
    ".inst 0x6f82e395  // udot v21.4s, v28.16b, v2.4b[0]\n"
    "movi v20.4s, #0x0\n"
    "movi v29.4s, #0x1\n"
    ".inst 0x6f82eb94  // udot v20.4s, v28.16b, v2.4b[2]\n"
    "add x20, %x[qp], %[offsetof_Requantize32_c_offset]\n"
    "ld1r { v14.4s }, [x20]\n"
    "ext v16.16b, v16.16b, v16.16b, #0x1\n"
    "zip1 v1.2d, v1.2d, v26.2d\n"
    ".inst 0x6fa3e3b2  // udot v18.4s, v29.16b, v3.4b[1]\n"
    "zip1 v5.2d, v5.2d, v13.2d\n"
    "zip1 v6.2d, v6.2d, v19.2d\n"
    ".inst 0x6fa3ebbf  // udot v31.4s, v29.16b, v3.4b[3]\n"
    "add x20, %x[qp], %[offsetof_Requantize32_minval]\n"
    "ld1r { v13.4s }, [x20]\n"
    "zip1 v7.2d, v7.2d, v17.2d\n"
    "movi v22.4s, #0x0\n"
    ".inst 0x6fa4e3b8  // udot v24.4s, v29.16b, v4.4b[1]\n"
    "movi v26.4s, #0x0\n"
    ".inst 0x6fa4ebbe  // udot v30.4s, v29.16b, v4.4b[3]\n"
    ".inst 0x6f81e396  // udot v22.4s, v28.16b, v1.4b[0]\n"
    "add x20, %x[qp], %[offsetof_Requantize32_maxval]\n"
    "ld1r { v15.4s }, [x20]\n"
    "movi v25.4s, #0x0\n"
    "movi v27.4s, #0x0\n"
    ".inst 0x6f81eb9a  // udot v26.4s, v28.16b, v1.4b[2]\n"
    "zip1 v0.2d, v0.2d, v16.2d\n"
    "movi v19.4s, #0x0\n"
    ".inst 0x6f85e399  // udot v25.4s, v28.16b, v5.4b[0]\n"
    "cmp %x[n_channels], #0x4\n"
    ".inst 0x6f85eb9b  // udot v27.4s, v28.16b, v5.4b[2]\n"
    ".inst 0x6f86e393  // udot v19.4s, v28.16b, v6.4b[0]\n"
    "add v24.4s, v18.4s, v24.4s\n"
    "mov x9, #0x0\n"
    "movi v18.4s, #0x0\n"
    ".inst 0x6f86eb92  // udot v18.4s, v28.16b, v6.4b[2]\n"
    ".inst 0x6fa2e3b5  // udot v21.4s, v29.16b, v2.4b[1]\n"
    "mov x28, #0x0\n"
    ".inst 0x6fa2ebb4  // udot v20.4s, v29.16b, v2.4b[3]\n"
    "add v17.4s, v31.4s, v30.4s\n"
    ".inst 0x6fa1e3b6  // udot v22.4s, v29.16b, v1.4b[1]\n"
    "ldp x27, x26, [%x[outptrs], #0x0]\n"
    "movi v16.4s, #0x0\n"
    ".inst 0x6f87e390  // udot v16.4s, v28.16b, v7.4b[0]\n"
    ".inst 0x6fa1ebba  // udot v26.4s, v29.16b, v1.4b[3]\n"
    "ldp x25, x24, [%x[outptrs], #0x10]\n"
    ".inst 0x6fa5e3b9  // udot v25.4s, v29.16b, v5.4b[1]\n"
    ".inst 0x6fa5ebbb  // udot v27.4s, v29.16b, v5.4b[3]\n"
    "add v30.4s, v21.4s, v24.4s\n"
    "ldp x23, x22, [%x[outptrs], #0x20]\n"
    ".inst 0x6fa6e3b3  // udot v19.4s, v29.16b, v6.4b[1]\n"
    ".inst 0x6fa6ebb2  // udot v18.4s, v29.16b, v6.4b[3]\n"
    "add v31.4s, v20.4s, v17.4s\n"
    "ldp x21, x20, [%x[outptrs], #0x30]\n"
    ".inst 0x6fa7e3b0  // udot v16.4s, v29.16b, v7.4b[1]\n"
    "add v22.4s, v22.4s, v30.4s\n"
    "add %x[params], %x[params], #0x50\n"
    "add v21.4s, v26.4s, v31.4s\n"
    "add v20.4s, v25.4s, v19.4s\n"
    "add v19.4s, v27.4s, v18.4s\n"
    "add v18.4s, v16.4s, v24.4s\n"
    "movi v16.4s, #0x0\n"
    ".inst 0x6f87eb90  // udot v16.4s, v28.16b, v7.4b[2]\n"
    ".inst 0x6fa7ebb0  // udot v16.4s, v29.16b, v7.4b[3]\n"
    "add v17.4s, v16.4s, v17.4s\n"
    "movi v16.4s, #0x0\n"
    ".inst 0x6f80e390  // udot v16.4s, v28.16b, v0.4b[0]\n"
    ".inst 0x6fa0e3b0  // udot v16.4s, v29.16b, v0.4b[1]\n"
    "add v24.4s, v22.4s, v16.4s\n"
    "add v26.4s, v22.4s, v25.4s\n"
    "movi v16.4s, #0x0\n"
    ".inst 0x6f80eb90  // udot v16.4s, v28.16b, v0.4b[2]\n"
    ".inst 0x6fa0ebb0  // udot v16.4s, v29.16b, v0.4b[3]\n"
    "add v25.4s, v21.4s, v16.4s\n"
    "add v27.4s, v21.4s, v27.4s\n"
    "add v28.4s, v20.4s, v30.4s\n"
    "add v29.4s, v19.4s, v31.4s\n"
    "add v30.4s, v18.4s, v20.4s\n"
    "add v31.4s, v17.4s, v19.4s\n"
    "neg v23.4s, v23.4s\n"
    "mul v24.4s, v24.4s, v23.4s\n"
    "mul v25.4s, v25.4s, v23.4s\n"
    "mul v26.4s, v26.4s, v23.4s\n"
    "mul v27.4s, v27.4s, v23.4s\n"
    "mul v28.4s, v28.4s, v23.4s\n"
    "mul v29.4s, v29.4s, v23.4s\n"
    "mul v30.4s, v30.4s, v23.4s\n"
    "mul v31.4s, v31.4s, v23.4s\n"
    "zip1 v19.4s, v24.4s, v26.4s\n"
    "zip1 v18.4s, v25.4s, v27.4s\n"
    "zip1 v17.4s, v28.4s, v30.4s\n"
    "zip1 v16.4s, v29.4s, v31.4s\n"
    "zip1 v22.4s, v19.4s, v18.4s\n"
    "zip1 v23.4s, v17.4s, v16.4s\n"
    "add v24.4s, v24.4s, v12.4s\n"
    "add v25.4s, v25.4s, v12.4s\n"
    "add v26.4s, v26.4s, v12.4s\n"
    "add v27.4s, v27.4s, v12.4s\n"
    "add v28.4s, v28.4s, v12.4s\n"
    "add v29.4s, v29.4s, v12.4s\n"
    "add v30.4s, v30.4s, v12.4s\n"
    "add v31.4s, v31.4s, v12.4s\n"
    "ble 2f\n"
    "1:"  // Loop
    "ldr q21, [%x[params], #0x60]\n"
    "ldr q20, [%x[params], #0x70]\n"
    ".inst 0x6f80e118  // udot v24.4s, v8.16b, v0.4b[0]\n"
    ".inst 0x6f80e919  // udot v25.4s, v8.16b, v0.4b[2]\n"
    "ldr q12, [%x[params], #0x80]\n"
    ".inst 0x6f81e11a  // udot v26.4s, v8.16b, v1.4b[0]\n"
    ".inst 0x6f81e91b  // udot v27.4s, v8.16b, v1.4b[2]\n"
    "sub %x[n_channels], %x[n_channels], #0x4\n"
    ".inst 0x6fa0e138  // udot v24.4s, v9.16b, v0.4b[1]\n"
    ".inst 0x6fa0e939  // udot v25.4s, v9.16b, v0.4b[3]\n"
    "cmp %x[n_channels], #0x4\n"
    "add x9, x9, #0x10\n"
    ".inst 0x6fa1e13a  // udot v26.4s, v9.16b, v1.4b[1]\n"
    ".inst 0x6fa1e93b  // udot v27.4s, v9.16b, v1.4b[3]\n"
    ".inst 0x6f82e11c  // udot v28.4s, v8.16b, v2.4b[0]\n"
    ".inst 0x6f82e91d  // udot v29.4s, v8.16b, v2.4b[2]\n"
    ".inst 0x6f83e11e  // udot v30.4s, v8.16b, v3.4b[0]\n"
    ".inst 0x6f83e91f  // udot v31.4s, v8.16b, v3.4b[2]\n"
    "ldr q8, [%x[params], #0x0]\n"
    ".inst 0x6f81e158  // udot v24.4s, v10.16b, v1.4b[0]\n"
    ".inst 0x6f81e959  // udot v25.4s, v10.16b, v1.4b[2]\n"
    ".inst 0x6f82e15a  // udot v26.4s, v10.16b, v2.4b[0]\n"
    ".inst 0x6f82e95b  // udot v27.4s, v10.16b, v2.4b[2]\n"
    ".inst 0x6fa2e13c  // udot v28.4s, v9.16b, v2.4b[1]\n"
    ".inst 0x6fa2e93d  // udot v29.4s, v9.16b, v2.4b[3]\n"
    ".inst 0x6fa3e13e  // udot v30.4s, v9.16b, v3.4b[1]\n"
    ".inst 0x6fa3e93f  // udot v31.4s, v9.16b, v3.4b[3]\n"
    "ldr q9, [%x[params], #0x10]\n"
    ".inst 0x6fa1e178  // udot v24.4s, v11.16b, v1.4b[1]\n"
    ".inst 0x6fa1e979  // udot v25.4s, v11.16b, v1.4b[3]\n"
    ".inst 0x6fa2e17a  // udot v26.4s, v11.16b, v2.4b[1]\n"
    ".inst 0x6fa2e97b  // udot v27.4s, v11.16b, v2.4b[3]\n"
    ".inst 0x6f83e15c  // udot v28.4s, v10.16b, v3.4b[0]\n"
    ".inst 0x6f83e95d  // udot v29.4s, v10.16b, v3.4b[2]\n"
    ".inst 0x6f84e15e  // udot v30.4s, v10.16b, v4.4b[0]\n"
    ".inst 0x6f84e95f  // udot v31.4s, v10.16b, v4.4b[2]\n"
    "ldr q10, [%x[params], #0x20]\n"
    ".inst 0x6f82e118  // udot v24.4s, v8.16b, v2.4b[0]\n"
    ".inst 0x6f82e919  // udot v25.4s, v8.16b, v2.4b[2]\n"
    ".inst 0x6f83e11a  // udot v26.4s, v8.16b, v3.4b[0]\n"
    ".inst 0x6f83e91b  // udot v27.4s, v8.16b, v3.4b[2]\n"
    ".inst 0x6fa3e17c  // udot v28.4s, v11.16b, v3.4b[1]\n"
    ".inst 0x6fa3e97d  // udot v29.4s, v11.16b, v3.4b[3]\n"
    ".inst 0x6fa4e17e  // udot v30.4s, v11.16b, v4.4b[1]\n"
    ".inst 0x6fa4e97f  // udot v31.4s, v11.16b, v4.4b[3]\n"
    "ldr q11, [%x[params], #0x30]\n"
    ".inst 0x6fa2e138  // udot v24.4s, v9.16b, v2.4b[1]\n"
    ".inst 0x6fa2e939  // udot v25.4s, v9.16b, v2.4b[3]\n"
    ".inst 0x6fa3e13a  // udot v26.4s, v9.16b, v3.4b[1]\n"
    ".inst 0x6fa3e93b  // udot v27.4s, v9.16b, v3.4b[3]\n"
    ".inst 0x6f84e11c  // udot v28.4s, v8.16b, v4.4b[0]\n"
    ".inst 0x6f84e91d  // udot v29.4s, v8.16b, v4.4b[2]\n"
    ".inst 0x6f85e11e  // udot v30.4s, v8.16b, v5.4b[0]\n"
    ".inst 0x6f85e91f  // udot v31.4s, v8.16b, v5.4b[2]\n"
    "ldr q8, [%x[params], #0x40]\n"
    ".inst 0x6f83e158  // udot v24.4s, v10.16b, v3.4b[0]\n"
    ".inst 0x6f83e959  // udot v25.4s, v10.16b, v3.4b[2]\n"
    ".inst 0x6f84e15a  // udot v26.4s, v10.16b, v4.4b[0]\n"
    ".inst 0x6f84e95b  // udot v27.4s, v10.16b, v4.4b[2]\n"
    ".inst 0x6fa4e13c  // udot v28.4s, v9.16b, v4.4b[1]\n"
    ".inst 0x6fa4e93d  // udot v29.4s, v9.16b, v4.4b[3]\n"
    ".inst 0x6fa5e13e  // udot v30.4s, v9.16b, v5.4b[1]\n"
    ".inst 0x6fa5e93f  // udot v31.4s, v9.16b, v5.4b[3]\n"
    "ldr q9, [%x[params], #0x50]\n"
    ".inst 0x6fa3e178  // udot v24.4s, v11.16b, v3.4b[1]\n"
    ".inst 0x6fa3e979  // udot v25.4s, v11.16b, v3.4b[3]\n"
    ".inst 0x6fa4e17a  // udot v26.4s, v11.16b, v4.4b[1]\n"
    ".inst 0x6fa4e97b  // udot v27.4s, v11.16b, v4.4b[3]\n"
    ".inst 0x6f85e15c  // udot v28.4s, v10.16b, v5.4b[0]\n"
    ".inst 0x6f85e95d  // udot v29.4s, v10.16b, v5.4b[2]\n"
    ".inst 0x6f86e15e  // udot v30.4s, v10.16b, v6.4b[0]\n"
    ".inst 0x6f86e95f  // udot v31.4s, v10.16b, v6.4b[2]\n"
    "ldr q10, [%x[params], #0xb0]\n"
    ".inst 0x6f84e118  // udot v24.4s, v8.16b, v4.4b[0]\n"
    ".inst 0x6f84e919  // udot v25.4s, v8.16b, v4.4b[2]\n"
    ".inst 0x6f85e11a  // udot v26.4s, v8.16b, v5.4b[0]\n"
    ".inst 0x6f85e91b  // udot v27.4s, v8.16b, v5.4b[2]\n"
    ".inst 0x6fa5e17c  // udot v28.4s, v11.16b, v5.4b[1]\n"
    ".inst 0x6fa5e97d  // udot v29.4s, v11.16b, v5.4b[3]\n"
    ".inst 0x6fa6e17e  // udot v30.4s, v11.16b, v6.4b[1]\n"
    ".inst 0x6fa6e97f  // udot v31.4s, v11.16b, v6.4b[3]\n"
    "ldr q11, [%x[params], #0xc0]\n"
    ".inst 0x6fa4e138  // udot v24.4s, v9.16b, v4.4b[1]\n"
    ".inst 0x6fa4e939  // udot v25.4s, v9.16b, v4.4b[3]\n"
    "sqrdmulh v24.4s, v24.4s, v21.4s\n"
    ".inst 0x6fa5e13a  // udot v26.4s, v9.16b, v5.4b[1]\n"
    ".inst 0x6fa5e93b  // udot v27.4s, v9.16b, v5.4b[3]\n"
    "sqrdmulh v25.4s, v25.4s, v21.4s\n"
    ".inst 0x6f86e11c  // udot v28.4s, v8.16b, v6.4b[0]\n"
    ".inst 0x6f86e91d  // udot v29.4s, v8.16b, v6.4b[2]\n"
    "sqrdmulh v26.4s, v26.4s, v21.4s\n"
    ".inst 0x6f87e11e  // udot v30.4s, v8.16b, v7.4b[0]\n"
    ".inst 0x6f87e91f  // udot v31.4s, v8.16b, v7.4b[2]\n"
    "ldr q8, [%x[params], #0x90]\n"
    "sqrdmulh v27.4s, v27.4s, v21.4s\n"
    ".inst 0x6fa6e13c  // udot v28.4s, v9.16b, v6.4b[1]\n"
    ".inst 0x6fa6e93d  // udot v29.4s, v9.16b, v6.4b[3]\n"
    "and v19.16b, v24.16b, v20.16b\n"
    ".inst 0x6fa7e13e  // udot v30.4s, v9.16b, v7.4b[1]\n"
    ".inst 0x6fa7e93f  // udot v31.4s, v9.16b, v7.4b[3]\n"
    "ldr q9, [%x[params], #0xa0]\n"
    "and v18.16b, v25.16b, v20.16b\n"
    "sshr v19.4s, v19.4s, #0x1f\n"
    "sshr v18.4s, v18.4s, #0x1f\n"
    "add %x[params], %x[params], #0xd0\n"
    "sqrdmulh v28.4s, v28.4s, v21.4s\n"
    "sqrdmulh v29.4s, v29.4s, v21.4s\n"
    "sqrdmulh v30.4s, v30.4s, v21.4s\n"
    "sqrdmulh v31.4s, v31.4s, v21.4s\n"
    "and v17.16b, v26.16b, v20.16b\n"
    "sshr v17.4s, v17.4s, #0x1f\n"
    "sqadd v24.4s, v24.4s, v19.4s\n"
    "and v16.16b, v27.16b, v20.16b\n"
    "sshr v16.4s, v16.4s, #0x1f\n"
    "sqadd v25.4s, v25.4s, v18.4s\n"
    "sqadd v26.4s, v26.4s, v17.4s\n"
    "sqadd v27.4s, v27.4s, v16.4s\n"
    "and v19.16b, v28.16b, v20.16b\n"
    "and v18.16b, v29.16b, v20.16b\n"
    "and v17.16b, v30.16b, v20.16b\n"
    "sshr v19.4s, v19.4s, #0x1f\n"
    "sshr v18.4s, v18.4s, #0x1f\n"
    "sshr v17.4s, v17.4s, #0x1f\n"
    "sqadd v28.4s, v28.4s, v19.4s\n"
    "and v16.16b, v31.16b, v20.16b\n"
    "sshr v16.4s, v16.4s, #0x1f\n"
    "sqadd v29.4s, v29.4s, v18.4s\n"
    "sqadd v30.4s, v30.4s, v17.4s\n"
    "sqadd v31.4s, v31.4s, v16.4s\n"
    "srshl v24.4s, v24.4s, v20.4s\n"
    "srshl v25.4s, v25.4s, v20.4s\n"
    "srshl v26.4s, v26.4s, v20.4s\n"
    "srshl v27.4s, v27.4s, v20.4s\n"
    "srshl v28.4s, v28.4s, v20.4s\n"
    "srshl v29.4s, v29.4s, v20.4s\n"
    "srshl v30.4s, v30.4s, v20.4s\n"
    "srshl v31.4s, v31.4s, v20.4s\n"
    "add v24.4s, v24.4s, v14.4s\n"
    "add v25.4s, v25.4s, v14.4s\n"
    "add v26.4s, v26.4s, v14.4s\n"
    "add v27.4s, v27.4s, v14.4s\n"
    "add v28.4s, v28.4s, v14.4s\n"
    "add v29.4s, v29.4s, v14.4s\n"
    "add v30.4s, v30.4s, v14.4s\n"
    "add v31.4s, v31.4s, v14.4s\n"
    "smin v24.4s, v24.4s, v15.4s\n"
    "smin v25.4s, v25.4s, v15.4s\n"
    "smin v26.4s, v26.4s, v15.4s\n"
    "smin v27.4s, v27.4s, v15.4s\n"
    "smin v28.4s, v28.4s, v15.4s\n"
    "smin v29.4s, v29.4s, v15.4s\n"
    "smin v30.4s, v30.4s, v15.4s\n"
    "smin v31.4s, v31.4s, v15.4s\n"
    "smax v24.4s, v24.4s, v13.4s\n"
    "smax v25.4s, v25.4s, v13.4s\n"
    "smax v26.4s, v26.4s, v13.4s\n"
    "smax v27.4s, v27.4s, v13.4s\n"
    "smax v28.4s, v28.4s, v13.4s\n"
    "smax v29.4s, v29.4s, v13.4s\n"
    "smax v30.4s, v30.4s, v13.4s\n"
    "smax v31.4s, v31.4s, v13.4s\n"
    "uzp1 v24.16b, v24.16b, v24.16b\n"
    "uzp1 v25.16b, v25.16b, v25.16b\n"
    "uzp1 v26.16b, v26.16b, v26.16b\n"
    "uzp1 v27.16b, v27.16b, v27.16b\n"
    "uzp1 v28.16b, v28.16b, v28.16b\n"
    "uzp1 v29.16b, v29.16b, v29.16b\n"
    "uzp1 v30.16b, v30.16b, v30.16b\n"
    "uzp1 v31.16b, v31.16b, v31.16b\n"
    "uzp1 v24.16b, v24.16b, v24.16b\n"
    "uzp1 v25.16b, v25.16b, v25.16b\n"
    "str s24, [x27, x28]\n"
    "uzp1 v26.16b, v26.16b, v26.16b\n"
    "uzp1 v27.16b, v27.16b, v27.16b\n"
    "str s25, [x26, x28]\n"
    "uzp1 v28.16b, v28.16b, v28.16b\n"
    "uzp1 v29.16b, v29.16b, v29.16b\n"
    "str s26, [x25, x28]\n"
    "uzp1 v30.16b, v30.16b, v30.16b\n"
    "uzp1 v31.16b, v31.16b, v31.16b\n"
    "str s27, [x24, x28]\n"
    "str s28, [x23, x28]\n"
    "dup v24.4s, v22.s[0]\n"
    "dup v25.4s, v22.s[1]\n"
    "str s29, [x22, x28]\n"
    "dup v26.4s, v22.s[2]\n"
    "dup v27.4s, v22.s[3]\n"
    "str s30, [x21, x28]\n"
    "dup v28.4s, v23.s[0]\n"
    "dup v29.4s, v23.s[1]\n"
    "str s31, [x20, x28]\n"
    "dup v30.4s, v23.s[2]\n"
    "dup v31.4s, v23.s[3]\n"
    "add x28, x28, #0x4\n"
    "add v24.4s, v24.4s, v12.4s\n"
    "add v25.4s, v25.4s, v12.4s\n"
    "add v26.4s, v26.4s, v12.4s\n"
    "add v27.4s, v27.4s, v12.4s\n"
    "add v28.4s, v28.4s, v12.4s\n"
    "add v29.4s, v29.4s, v12.4s\n"
    "add v30.4s, v30.4s, v12.4s\n"
    "add v31.4s, v31.4s, v12.4s\n"
    "bgt 1b\n"
    "2:"  // Tail
    "ldr q21, [%x[params], #0x60]\n"
    "ldr q20, [%x[params], #0x70]\n"
    ".inst 0x6f80e118  // udot v24.4s, v8.16b, v0.4b[0]\n"
    ".inst 0x6f80e919  // udot v25.4s, v8.16b, v0.4b[2]\n"
    ".inst 0x6f81e11a  // udot v26.4s, v8.16b, v1.4b[0]\n"
    ".inst 0x6f81e91b  // udot v27.4s, v8.16b, v1.4b[2]\n"
    "cmp %x[n_channels], #0x4\n"
    "add x27, x27, x28\n"
    ".inst 0x6fa0e138  // udot v24.4s, v9.16b, v0.4b[1]\n"
    ".inst 0x6fa0e939  // udot v25.4s, v9.16b, v0.4b[3]\n"
    "add x26, x26, x28\n"
    "add x25, x25, x28\n"
    ".inst 0x6fa1e13a  // udot v26.4s, v9.16b, v1.4b[1]\n"
    ".inst 0x6fa1e93b  // udot v27.4s, v9.16b, v1.4b[3]\n"
    "add x24, x24, x28\n"
    "add x23, x23, x28\n"
    ".inst 0x6f82e11c  // udot v28.4s, v8.16b, v2.4b[0]\n"
    ".inst 0x6f82e91d  // udot v29.4s, v8.16b, v2.4b[2]\n"
    "add x22, x22, x28\n"
    "add x21, x21, x28\n"
    ".inst 0x6f83e11e  // udot v30.4s, v8.16b, v3.4b[0]\n"
    ".inst 0x6f83e91f  // udot v31.4s, v8.16b, v3.4b[2]\n"
    "ldr q8, [%x[params], #0x0]\n"
    "add x20, x20, x28\n"
    ".inst 0x6f81e158  // udot v24.4s, v10.16b, v1.4b[0]\n"
    ".inst 0x6f81e959  // udot v25.4s, v10.16b, v1.4b[2]\n"
    ".inst 0x6f82e15a  // udot v26.4s, v10.16b, v2.4b[0]\n"
    ".inst 0x6f82e95b  // udot v27.4s, v10.16b, v2.4b[2]\n"
    ".inst 0x6fa2e13c  // udot v28.4s, v9.16b, v2.4b[1]\n"
    ".inst 0x6fa2e93d  // udot v29.4s, v9.16b, v2.4b[3]\n"
    ".inst 0x6fa3e13e  // udot v30.4s, v9.16b, v3.4b[1]\n"
    ".inst 0x6fa3e93f  // udot v31.4s, v9.16b, v3.4b[3]\n"
    "ldr q9, [%x[params], #0x10]\n"
    ".inst 0x6fa1e178  // udot v24.4s, v11.16b, v1.4b[1]\n"
    ".inst 0x6fa1e979  // udot v25.4s, v11.16b, v1.4b[3]\n"
    ".inst 0x6fa2e17a  // udot v26.4s, v11.16b, v2.4b[1]\n"
    ".inst 0x6fa2e97b  // udot v27.4s, v11.16b, v2.4b[3]\n"
    ".inst 0x6f83e15c  // udot v28.4s, v10.16b, v3.4b[0]\n"
    ".inst 0x6f83e95d  // udot v29.4s, v10.16b, v3.4b[2]\n"
    ".inst 0x6f84e15e  // udot v30.4s, v10.16b, v4.4b[0]\n"
    ".inst 0x6f84e95f  // udot v31.4s, v10.16b, v4.4b[2]\n"
    "ldr q10, [%x[params], #0x20]\n"
    ".inst 0x6f82e118  // udot v24.4s, v8.16b, v2.4b[0]\n"
    ".inst 0x6f82e919  // udot v25.4s, v8.16b, v2.4b[2]\n"
    ".inst 0x6f83e11a  // udot v26.4s, v8.16b, v3.4b[0]\n"
    ".inst 0x6f83e91b  // udot v27.4s, v8.16b, v3.4b[2]\n"
    ".inst 0x6fa3e17c  // udot v28.4s, v11.16b, v3.4b[1]\n"
    ".inst 0x6fa3e97d  // udot v29.4s, v11.16b, v3.4b[3]\n"
    ".inst 0x6fa4e17e  // udot v30.4s, v11.16b, v4.4b[1]\n"
    ".inst 0x6fa4e97f  // udot v31.4s, v11.16b, v4.4b[3]\n"
    "ldr q11, [%x[params], #0x30]\n"
    ".inst 0x6fa2e138  // udot v24.4s, v9.16b, v2.4b[1]\n"
    ".inst 0x6fa2e939  // udot v25.4s, v9.16b, v2.4b[3]\n"
    ".inst 0x6fa3e13a  // udot v26.4s, v9.16b, v3.4b[1]\n"
    ".inst 0x6fa3e93b  // udot v27.4s, v9.16b, v3.4b[3]\n"
    ".inst 0x6f84e11c  // udot v28.4s, v8.16b, v4.4b[0]\n"
    ".inst 0x6f84e91d  // udot v29.4s, v8.16b, v4.4b[2]\n"
    ".inst 0x6f85e11e  // udot v30.4s, v8.16b, v5.4b[0]\n"
    ".inst 0x6f85e91f  // udot v31.4s, v8.16b, v5.4b[2]\n"
    "ldr q8, [%x[params], #0x40]\n"
    ".inst 0x6f83e158  // udot v24.4s, v10.16b, v3.4b[0]\n"
    ".inst 0x6f83e959  // udot v25.4s, v10.16b, v3.4b[2]\n"
    ".inst 0x6f84e15a  // udot v26.4s, v10.16b, v4.4b[0]\n"
    ".inst 0x6f84e95b  // udot v27.4s, v10.16b, v4.4b[2]\n"
    ".inst 0x6fa4e13c  // udot v28.4s, v9.16b, v4.4b[1]\n"
    ".inst 0x6fa4e93d  // udot v29.4s, v9.16b, v4.4b[3]\n"
    ".inst 0x6fa5e13e  // udot v30.4s, v9.16b, v5.4b[1]\n"
    ".inst 0x6fa5e93f  // udot v31.4s, v9.16b, v5.4b[3]\n"
    "ldr q9, [%x[params], #0x50]\n"
    "add %x[params], %x[params], #0x80\n"
    ".inst 0x6fa3e178  // udot v24.4s, v11.16b, v3.4b[1]\n"
    ".inst 0x6fa3e979  // udot v25.4s, v11.16b, v3.4b[3]\n"
    ".inst 0x6fa4e17a  // udot v26.4s, v11.16b, v4.4b[1]\n"
    ".inst 0x6fa4e97b  // udot v27.4s, v11.16b, v4.4b[3]\n"
    ".inst 0x6f85e15c  // udot v28.4s, v10.16b, v5.4b[0]\n"
    ".inst 0x6f85e95d  // udot v29.4s, v10.16b, v5.4b[2]\n"
    ".inst 0x6f86e15e  // udot v30.4s, v10.16b, v6.4b[0]\n"
    ".inst 0x6f86e95f  // udot v31.4s, v10.16b, v6.4b[2]\n"
    ".inst 0x6f84e118  // udot v24.4s, v8.16b, v4.4b[0]\n"
    ".inst 0x6f84e919  // udot v25.4s, v8.16b, v4.4b[2]\n"
    ".inst 0x6f85e11a  // udot v26.4s, v8.16b, v5.4b[0]\n"
    ".inst 0x6f85e91b  // udot v27.4s, v8.16b, v5.4b[2]\n"
    ".inst 0x6fa5e17c  // udot v28.4s, v11.16b, v5.4b[1]\n"
    ".inst 0x6fa5e97d  // udot v29.4s, v11.16b, v5.4b[3]\n"
    ".inst 0x6fa6e17e  // udot v30.4s, v11.16b, v6.4b[1]\n"
    ".inst 0x6fa6e97f  // udot v31.4s, v11.16b, v6.4b[3]\n"
    ".inst 0x6fa4e138  // udot v24.4s, v9.16b, v4.4b[1]\n"
    ".inst 0x6fa4e939  // udot v25.4s, v9.16b, v4.4b[3]\n"
    "sqrdmulh v24.4s, v24.4s, v21.4s\n"
    ".inst 0x6fa5e13a  // udot v26.4s, v9.16b, v5.4b[1]\n"
    ".inst 0x6fa5e93b  // udot v27.4s, v9.16b, v5.4b[3]\n"
    "sqrdmulh v25.4s, v25.4s, v21.4s\n"
    ".inst 0x6f86e11c  // udot v28.4s, v8.16b, v6.4b[0]\n"
    ".inst 0x6f86e91d  // udot v29.4s, v8.16b, v6.4b[2]\n"
    "sqrdmulh v26.4s, v26.4s, v21.4s\n"
    ".inst 0x6f87e11e  // udot v30.4s, v8.16b, v7.4b[0]\n"
    ".inst 0x6f87e91f  // udot v31.4s, v8.16b, v7.4b[2]\n"
    "sqrdmulh v27.4s, v27.4s, v21.4s\n"
    ".inst 0x6fa6e13c  // udot v28.4s, v9.16b, v6.4b[1]\n"
    ".inst 0x6fa6e93d  // udot v29.4s, v9.16b, v6.4b[3]\n"
    "and v19.16b, v24.16b, v20.16b\n"
    ".inst 0x6fa7e13e  // udot v30.4s, v9.16b, v7.4b[1]\n"
    ".inst 0x6fa7e93f  // udot v31.4s, v9.16b, v7.4b[3]\n"
    "and v18.16b, v25.16b, v20.16b\n"
    "and v17.16b, v26.16b, v20.16b\n"
    "and v16.16b, v27.16b, v20.16b\n"
    "sshr v19.4s, v19.4s, #0x1f\n"
    "sshr v18.4s, v18.4s, #0x1f\n"
    "sshr v17.4s, v17.4s, #0x1f\n"
    "sshr v16.4s, v16.4s, #0x1f\n"
    "sqrdmulh v28.4s, v28.4s, v21.4s\n"
    "sqrdmulh v29.4s, v29.4s, v21.4s\n"
    "sqrdmulh v30.4s, v30.4s, v21.4s\n"
    "sqrdmulh v31.4s, v31.4s, v21.4s\n"
    "sqadd v24.4s, v24.4s, v19.4s\n"
    "sqadd v25.4s, v25.4s, v18.4s\n"
    "sqadd v26.4s, v26.4s, v17.4s\n"
    "sqadd v27.4s, v27.4s, v16.4s\n"
    "and v19.16b, v28.16b, v20.16b\n"
    "and v18.16b, v29.16b, v20.16b\n"
    "and v17.16b, v30.16b, v20.16b\n"
    "and v16.16b, v31.16b, v20.16b\n"
    "sshr v19.4s, v19.4s, #0x1f\n"
    "sshr v18.4s, v18.4s, #0x1f\n"
    "sshr v17.4s, v17.4s, #0x1f\n"
    "sshr v16.4s, v16.4s, #0x1f\n"
    "sqadd v28.4s, v28.4s, v19.4s\n"
    "sqadd v29.4s, v29.4s, v18.4s\n"
    "sqadd v30.4s, v30.4s, v17.4s\n"
    "sqadd v31.4s, v31.4s, v16.4s\n"
    "srshl v24.4s, v24.4s, v20.4s\n"
    "srshl v25.4s, v25.4s, v20.4s\n"
    "srshl v26.4s, v26.4s, v20.4s\n"
    "srshl v27.4s, v27.4s, v20.4s\n"
    "srshl v28.4s, v28.4s, v20.4s\n"
    "srshl v29.4s, v29.4s, v20.4s\n"
    "srshl v30.4s, v30.4s, v20.4s\n"
    "srshl v31.4s, v31.4s, v20.4s\n"
    "add v24.4s, v24.4s, v14.4s\n"
    "add v25.4s, v25.4s, v14.4s\n"
    "add v26.4s, v26.4s, v14.4s\n"
    "add v27.4s, v27.4s, v14.4s\n"
    "add v28.4s, v28.4s, v14.4s\n"
    "add v29.4s, v29.4s, v14.4s\n"
    "add v30.4s, v30.4s, v14.4s\n"
    "add v31.4s, v31.4s, v14.4s\n"
    "smin v24.4s, v24.4s, v15.4s\n"
    "smin v25.4s, v25.4s, v15.4s\n"
    "smin v26.4s, v26.4s, v15.4s\n"
    "smin v27.4s, v27.4s, v15.4s\n"
    "smin v28.4s, v28.4s, v15.4s\n"
    "smin v29.4s, v29.4s, v15.4s\n"
    "smin v30.4s, v30.4s, v15.4s\n"
    "smin v31.4s, v31.4s, v15.4s\n"
    "smax v24.4s, v24.4s, v13.4s\n"
    "smax v25.4s, v25.4s, v13.4s\n"
    "smax v26.4s, v26.4s, v13.4s\n"
    "smax v27.4s, v27.4s, v13.4s\n"
    "smax v28.4s, v28.4s, v13.4s\n"
    "smax v29.4s, v29.4s, v13.4s\n"
    "smax v30.4s, v30.4s, v13.4s\n"
    "smax v31.4s, v31.4s, v13.4s\n"
    "uzp1 v24.16b, v24.16b, v24.16b\n"
    "uzp1 v25.16b, v25.16b, v25.16b\n"
    "uzp1 v26.16b, v26.16b, v26.16b\n"
    "uzp1 v27.16b, v27.16b, v27.16b\n"
    "uzp1 v28.16b, v28.16b, v28.16b\n"
    "uzp1 v29.16b, v29.16b, v29.16b\n"
    "uzp1 v30.16b, v30.16b, v30.16b\n"
    "uzp1 v31.16b, v31.16b, v31.16b\n"
    "uzp1 v24.16b, v24.16b, v24.16b\n"
    "uzp1 v25.16b, v25.16b, v25.16b\n"
    "uzp1 v26.16b, v26.16b, v26.16b\n"
    "uzp1 v27.16b, v27.16b, v27.16b\n"
    "uzp1 v28.16b, v28.16b, v28.16b\n"
    "uzp1 v29.16b, v29.16b, v29.16b\n"
    "uzp1 v30.16b, v30.16b, v30.16b\n"
    "uzp1 v31.16b, v31.16b, v31.16b\n"
    "blt 3f\n"
    "str s24, [x27, #0x0]\n"
    "str s25, [x26, #0x0]\n"
    "str s26, [x25, #0x0]\n"
    "str s27, [x24, #0x0]\n"
    "str s28, [x23, #0x0]\n"
    "str s29, [x22, #0x0]\n"
    "str s30, [x21, #0x0]\n"
    "str s31, [x20, #0x0]\n"
    "b 4f\n"
    "3:"  // Tail: Oddments
    "subs %x[n_channels], %x[n_channels], #0x1\n"
    "st1 { v24.b }[0], [x27], #0x1\n"
    "st1 { v25.b }[0], [x26], #0x1\n"
    "st1 { v26.b }[0], [x25], #0x1\n"
    "st1 { v27.b }[0], [x24], #0x1\n"
    "st1 { v28.b }[0], [x23], #0x1\n"
    "st1 { v29.b }[0], [x22], #0x1\n"
    "st1 { v30.b }[0], [x21], #0x1\n"
    "st1 { v31.b }[0], [x20], #0x1\n"
    "beq 4f\n"
    "subs %x[n_channels], %x[n_channels], #0x1\n"
    "st1 { v24.b }[1], [x27], #0x1\n"
    "st1 { v25.b }[1], [x26], #0x1\n"
    "st1 { v26.b }[1], [x25], #0x1\n"
    "st1 { v27.b }[1], [x24], #0x1\n"
    "st1 { v28.b }[1], [x23], #0x1\n"
    "st1 { v29.b }[1], [x22], #0x1\n"
    "st1 { v30.b }[1], [x21], #0x1\n"
    "st1 { v31.b }[1], [x20], #0x1\n"
    "beq 4f\n"
    "subs %x[n_channels], %x[n_channels], #0x1\n"
    "st1 { v24.b }[2], [x27], #0x1\n"
    "st1 { v25.b }[2], [x26], #0x1\n"
    "st1 { v26.b }[2], [x25], #0x1\n"
    "st1 { v27.b }[2], [x24], #0x1\n"
    "st1 { v28.b }[2], [x23], #0x1\n"
    "st1 { v29.b }[2], [x22], #0x1\n"
    "st1 { v30.b }[2], [x21], #0x1\n"
    "st1 { v31.b }[2], [x20], #0x1\n"
    "beq 4f\n"
    "st1 { v24.b }[3], [x27], #0x1\n"
    "subs %x[n_channels], %x[n_channels], #0x1\n"
    "st1 { v25.b }[3], [x26], #0x1\n"
    "st1 { v26.b }[3], [x25], #0x1\n"
    "st1 { v27.b }[3], [x24], #0x1\n"
    "st1 { v28.b }[3], [x23], #0x1\n"
    "st1 { v29.b }[3], [x22], #0x1\n"
    "st1 { v30.b }[3], [x21], #0x1\n"
    "st1 { v31.b }[3], [x20], #0x1\n"
    "4:"  // Tail: End
    : [n_channels] "+&r" (n_output_channels), [params] "+&r" (params)
    : [inptrs] "r" (inptrs), [offsetof_Requantize32_b_offset] "I" (offsetof(arm_gemm::Requantize32, b_offset)), [offsetof_Requantize32_c_offset] "I" (offsetof(arm_gemm::Requantize32, c_offset)), [offsetof_Requantize32_maxval] "I" (offsetof(arm_gemm::Requantize32, maxval)), [offsetof_Requantize32_minval] "I" (offsetof(arm_gemm::Requantize32, minval)), [outptrs] "r" (outptrs), [qp] "r" (&qp)
    : "cc", "memory", "v0", "v1", "v2", "v3", "v4", "v5", "v6", "v7", "v8", "v9", "v10", "v11", "v12", "v13", "v14", "v15", "v16", "v17", "v18", "v19", "v20", "v21", "v22", "v23", "v24", "v25", "v26", "v27", "v28", "v29", "v30", "v31", "x9", "x20", "x21", "x22", "x23", "x24", "x25", "x26", "x27", "x28"
  );
}

}  // namespace depthwise
}  // namespace arm_conv
#endif // defined(__aarch64__)
